#include "pch.h"
#include "bitmap.h"
 
HPALETTE CreateDIBPalette (BITMAPINFO* a_pbmi, LPINT a_piNumColors)
{
	LPBITMAPINFOHEADER  lpbi;
	LPLOGPALETTE     lpPal;
	HANDLE           hLogPal;
	HPALETTE         hPal = NULL;
	int              i;
	
	lpbi = (LPBITMAPINFOHEADER)a_pbmi;
	if (lpbi->biBitCount <= 8)
	{
		*a_piNumColors = (1 << lpbi->biBitCount);
	}
	else
	{
		*a_piNumColors = 0;  // No palette needed for 16, 24 or 32 BPP DIB
	}
	
	if (*a_piNumColors)
	{
		hLogPal = GlobalAlloc (GHND, sizeof (LOGPALETTE) +
			sizeof (PALETTEENTRY) * (*a_piNumColors));
		lpPal = (LPLOGPALETTE) GlobalLock (hLogPal);
		lpPal->palVersion    = 0x300;
		lpPal->palNumEntries = *a_piNumColors;
		
		for (i = 0;  i < *a_piNumColors;  i++)
		{
			lpPal->palPalEntry[i].peRed   = a_pbmi->bmiColors[i].rgbRed;
			lpPal->palPalEntry[i].peGreen = a_pbmi->bmiColors[i].rgbGreen;
			lpPal->palPalEntry[i].peBlue  = a_pbmi->bmiColors[i].rgbBlue;
			lpPal->palPalEntry[i].peFlags = 0;
		}
		hPal = CreatePalette (lpPal);
		GlobalUnlock (hLogPal);
		GlobalFree   (hLogPal);
	}
	return hPal;
}

HBITMAP LoadBitmapResource(
	HINSTANCE	a_hInst,
	LPCSTR		a_pszResource,
	HPALETTE	*a_phPalette)
{
    HRSRC  hRsrc;
    HGLOBAL hGlobal;
    HBITMAP hBitmapFinal = NULL;
    LPBITMAPINFOHEADER  lpbi;
    HDC hdc;
    int iNumColors;

	// look for resources
    if (hRsrc = FindResource(a_hInst, a_pszResource, RT_BITMAP))
	{
		hGlobal = LoadResource(a_hInst, hRsrc);
		lpbi = (LPBITMAPINFOHEADER)LockResource(hGlobal);
		
		hdc = GetDC(NULL);
		*a_phPalette =  CreateDIBPalette ((LPBITMAPINFO)lpbi, &iNumColors);
		if (*a_phPalette)
		{
			SelectPalette(hdc,*a_phPalette,FALSE);
			RealizePalette(hdc);
		}
		
		hBitmapFinal = CreateDIBitmap(
			hdc,
			(LPBITMAPINFOHEADER)lpbi,
			(LONG)CBM_INIT,
			(LPSTR)lpbi + lpbi->biSize + iNumColors * sizeof(RGBQUAD),
			(LPBITMAPINFO)lpbi,
			DIB_RGB_COLORS );
		
		ReleaseDC(NULL,hdc);
		UnlockResource(hGlobal);
		FreeResource(hGlobal);
	}
    return (hBitmapFinal);
}

HBITMAP LoadBitmapFile(
	LPCSTR		a_pszFile,
	HPALETTE	*a_phPalette)
{
    void* hMem;
    HBITMAP hBitmapFinal = NULL;
    BITMAPINFOHEADER*	lpbi;
	BITMAPFILEHEADER*	lpbfh;
    HDC hdc;
    int iNumColors;
	HFILE hFile;
	DWORD size;

	// try to read the whole file to memory
	hFile = _lopen(a_pszFile, OF_READ | OF_SHARE_DENY_WRITE);
	if (hFile == HFILE_ERROR)
	{
		return NULL;
	}

	size = _llseek(hFile, 0, FILE_END);
	_llseek(hFile, 0, FILE_BEGIN);

	hMem = (void*) GlobalAlloc(GMEM_FIXED, size);

	_lread(hFile, hMem, size);
	_lclose(hFile);
	
	lpbfh = (BITMAPFILEHEADER*) hMem;

	// check for valid bmp file
	if (lpbfh->bfType == 'B' || ('M' << 8))
	{
		lpbi = (BITMAPINFOHEADER*) ((DWORD) hMem + sizeof(BITMAPFILEHEADER));
		
		hdc = GetDC(NULL);
		*a_phPalette =  CreateDIBPalette ((LPBITMAPINFO)lpbi, &iNumColors);
		if (*a_phPalette)
		{
			SelectPalette(hdc,*a_phPalette,FALSE);
			RealizePalette(hdc);
		}
		
		hBitmapFinal = CreateDIBitmap(
			hdc,
			(LPBITMAPINFOHEADER)lpbi,
			(LONG)CBM_INIT,
			(LPSTR)lpbi + lpbi->biSize + iNumColors * sizeof(RGBQUAD),
			(LPBITMAPINFO)lpbi,
			DIB_RGB_COLORS );
		
		ReleaseDC(NULL,hdc);
	}

	GlobalFree(hMem);

    return (hBitmapFinal);
}

CDibSection::CDibSection()
{
	m_hBitmap = NULL;
	m_hOldBitmap = NULL;
	m_hDC = NULL;
	m_pData = NULL;
	m_dwLineWidth = 0;
}

CDibSection::~CDibSection()
{
	Free();
}

BOOL CDibSection::Create(int a_lWidth, int a_lHeight, int a_lBitCount)
{
	Free();

	if (a_lBitCount != 8 &&
		a_lBitCount != 16 &&
		a_lBitCount != 24 &&
		a_lBitCount != 32)
	{
		return FALSE;
	}

	DWORD size = 
		sizeof(BITMAPINFOHEADER) +
		(a_lBitCount == 8 ? sizeof(RGBQUAD) * 256 : 0);
	BITMAPINFO* pbi = (BITMAPINFO*)new char[size];

	memset(pbi, 0, size);
	pbi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	pbi->bmiHeader.biWidth = a_lWidth;
	pbi->bmiHeader.biHeight = a_lHeight;
	pbi->bmiHeader.biPlanes = 1;
	pbi->bmiHeader.biBitCount = a_lBitCount;
	pbi->bmiHeader.biCompression = BI_RGB;

	// make it a inverse DIB.
	if (pbi->bmiHeader.biHeight > 0)
	{
		pbi->bmiHeader.biHeight = -pbi->bmiHeader.biHeight;
	}

	m_hDC = ::CreateCompatibleDC(NULL);
	m_hBitmap = CreateDIBSection(
		m_hDC,
		pbi,
		DIB_RGB_COLORS,
		(void**) &m_pData,
		NULL,
		NULL);
	m_hOldBitmap = (HBITMAP) ::SelectObject(m_hDC, m_hBitmap);
	delete [] ((char*) pbi);

	if (m_hBitmap != NULL)
	{
		m_lWidth = a_lWidth;
		m_lHeight = a_lHeight;
		m_lBitCount = a_lBitCount;
		m_dwLineWidth = DWORDAlign(m_lWidth * m_lBitCount / 8);
		return TRUE;
	}
	else
	{
		Free();
		return FALSE;
	}
}

BOOL CDibSection::Free()
{
	if (m_hOldBitmap != NULL)
	{
		::SelectObject(m_hDC, m_hOldBitmap);
	}

	if (m_hDC != NULL)
	{
		::DeleteDC(m_hDC);
	}

	if (m_hBitmap != NULL)
	{
		::DeleteObject(m_hBitmap);
	}
	m_hBitmap = NULL;
	m_hOldBitmap = NULL;
	m_hDC = NULL;
	m_pData = NULL;
	m_dwLineWidth = 0;

	return TRUE;
}

BOOL CDibSection::SetColorTable(const RGBQUAD* a_prgb)
{
	SetDIBColorTable(m_hDC, 0, 256, a_prgb);
	return TRUE;
}

BOOL CDibSection::GetColorTable(RGBQUAD* a_prgb)
{
	GetDIBColorTable(m_hDC, 0, 256, a_prgb);
	return TRUE;
}

BOOL CDibSection::Load(LPCSTR a_pszPathName, HPALETTE *a_phPalette)
{
    void* hMem;
    BITMAPINFOHEADER*	lpbi;
	BITMAPFILEHEADER*	lpbfh;
	HFILE hFile;
	DWORD size;

	// try to read the whole file to memory
	hFile = _lopen(a_pszPathName, OF_READ | OF_SHARE_DENY_WRITE);
	if (hFile == HFILE_ERROR)
	{
		return FALSE;
	}

	size = _llseek(hFile, 0, FILE_END);
	_llseek(hFile, 0, FILE_BEGIN);

	hMem = (void*) GlobalAlloc(GMEM_FIXED, size);

	_lread(hFile, hMem, size);
	_lclose(hFile);
	
	lpbfh = (BITMAPFILEHEADER*) hMem;
	if (lpbfh->bfType == 'B' || ('M' << 8))
	{
		lpbi = (BITMAPINFOHEADER*) ((DWORD) hMem + sizeof(BITMAPFILEHEADER));
		Create(lpbi->biWidth, lpbi->biHeight, lpbi->biBitCount);

		// set the color table
		if (lpbi->biBitCount == 8)
		{
			SetColorTable(
				(RGBQUAD*) ((DWORD) hMem + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER)));
		}

		// calculate the size of BITMAPINFO
		DWORD size = sizeof(BITMAPINFOHEADER);
		if (lpbi->biClrUsed == 0 &&
			lpbi->biBitCount == 8)
		{
			size += sizeof(RGBQUAD) * 256;
		}
		else if (lpbi->biClrUsed != 0)
		{
			size += sizeof(RGBQUAD) * lpbi->biClrUsed;
		}

		// copy the pixel data. flip the bitmap at the same time
		PBYTE pSrc = (PBYTE) ((DWORD) hMem + sizeof(BITMAPFILEHEADER) + size);

		for (int j = 0; j < lpbi->biHeight; j ++)
		{
			memcpy(
				GetPointer() + j * GetLineWidth(),
				pSrc + (GetHeight() - 1 - j) * GetLineWidth(),
				GetLineWidth());
		}

		// create palette
		if (a_phPalette != NULL)
		{
			int iNumColors;

			*a_phPalette =  CreateDIBPalette((LPBITMAPINFO)lpbi, &iNumColors);
		}

		GlobalFree(hMem);
		return TRUE;
	}
	else
	{
		GlobalFree(hMem);
		return FALSE;
	}
}

BOOL CDibSection::Load(HINSTANCE a_hInst, LPCSTR a_pszResName, HPALETTE *a_phPalette)
{
    HRSRC  hRsrc;
    HGLOBAL hGlobal;
    LPBITMAPINFOHEADER  lpbi;
    int iNumColors;

	// look for resources
    if (hRsrc = FindResource(a_hInst, a_pszResName, RT_BITMAP))
	{
		hGlobal = LoadResource(a_hInst, hRsrc);
		lpbi = (LPBITMAPINFOHEADER)LockResource(hGlobal);
		Create(lpbi->biWidth, lpbi->biHeight, lpbi->biBitCount);

		// set the color table
		if (lpbi->biBitCount == 8)
		{
			SetColorTable(
				(RGBQUAD*) ((DWORD) lpbi + sizeof(BITMAPINFOHEADER)));
		}

		// calculate the size of BITMAPINFO
		DWORD size = sizeof(BITMAPINFOHEADER);
		if (lpbi->biClrUsed == 0 &&
			lpbi->biBitCount == 8)
		{
			size += sizeof(RGBQUAD) * 256;
		}
		else if (lpbi->biClrUsed != 0)
		{
			size += sizeof(RGBQUAD) * lpbi->biClrUsed;
		}

		// copy the pixel data. flip the bitmap at the same time
		PBYTE pSrc = (PBYTE) ((DWORD) lpbi + size);
		for (int j = 0; j < lpbi->biHeight; j ++)
		{
			memcpy(
				GetPointer() + j * GetLineWidth(),
				pSrc + (GetHeight() - 1 - j) * GetLineWidth(),
				GetLineWidth());
		}

		// create palette
		if (a_phPalette != NULL)
		{
			*a_phPalette =  CreateDIBPalette((LPBITMAPINFO)lpbi, &iNumColors);
		}

		UnlockResource(hGlobal);
		FreeResource(hGlobal);
		return TRUE;
	}
	else
	{
		UnlockResource(hGlobal);
		FreeResource(hGlobal);
		return FALSE;
	}
}
